// Code with Encoder:
// Test MD03a / Pololu motor with encoder
//http://forum.arduino.cc/index.php?topic=8652.0
// speed control (PI), V & I display
// Credits:
//   Dallaby   http://letsmakerobots.com/node/19558#comment-49685
//   Bill Porter  http://www.billporter.info/?p=286
//   bobbyorr (nice connection diagram) http://forum.pololu.com/viewtopic.php?f=15&t=1923
//   pin changes by ewh for the VNH5019
//   Changed pinb reading to read pin 11, should be BIT 4, was BIT 3, encodepin b - 4 Nov 2014
//   motor and chip example - Manual pid calc
//   http://forum.arduino.cc/index.php?topic=8652.0  the basic site
//   http://forum.arduino.cc/index.php/topic,8652.0.html
//   http://www.pololu.com/product/708  the driver chip
//   http://www.pololu.com/catalog/product/1443 the motor
/************************************************
REVIEW PIN to ARDUINO SETTINGS FOR BOTH DRIVERS!! 
Check code for pre-defined pins for both drivers
v-v-v-v-v-v-v-
4 Nov 2014 - ewh - Added the pin change interuppt vector code
                   this works for speed change, etc.
5 Nov 2014 - ewh - Added code for direct pin setting
7 Nov 2014 - ewh - Revising PID settings and current variable
************************************************************/

#include <DualVNH5019MotorShield.h>
#include <PID_v1.h>
#include <PID_AutoTune_v0.h>

DualVNH5019MotorShield md; // default pins ok for this

//define variables we will be connecting to
double Setpoint, Input, Output;

PID myPID(&Input, &Output, &Setpoint, .4,1,1, DIRECT);


#define InA1            2  //10                // INA motor pin
#define InB1            4  //11                // INB motor pin 
#define PWM1            9  //6                 // PWM motor pin			       		
#define encodPinA1      11//10 //3                 // encoder A pin
#define encodPinB1      10//11                     // encoder B pin
#define Vpin            A3 //0    A2 for pot   // battery monitoring analog pin
#define Apin            A0 //1                 // motor current monitoring analog pin

#define CURRENT_LIMIT   3000                    // high current warning
#define LOW_BAT         8000 //10000            // low bat warning
#define LOOPTIME        100                     // PID loop time
#define NUMREADINGS     10                      // samples for Amp average

int readings[NUMREADINGS];
//int readings[] = {0,0,0,0,0,0,0,0,0,0};
unsigned long lastMilli = 0;                    // loop timing 
unsigned long lastMilliPrint = 0;               // loop timing
int speed_req = 50;                            // speed (Set Point)  
int speed_act = 0;                              // speed (actual value)
int PWM_val = 0;                                // (25% = 64; 50% = 127; 75% = 191; 100% = 255)
int voltage = 0;                                // in mV
int current = 0;                                // in mA
volatile long count = 0;                        // rev counter
float Kp =   .4;                                // PID proportional control Gain
float Kd =    1;                                // PID Derivitave control gain


void setup() {
  analogReference(EXTERNAL);         // Current external ref is 3.3V, appiled to the aREF pin
  Serial.begin(115200);
  pinMode(InA1, OUTPUT);
  pinMode(InB1, OUTPUT);
  pinMode(PWM1, OUTPUT);
  cli();  // disable interrupts    
    DDRB  |= (0 << 2);  //  Enable pin 10 as INPUT for encodPinA1 set low
    PORTB |= (1 << 2);  //  Set Arduino pin 10 HIGH for pull up (encodPinA1)
    DDRB  |= (0 << 3);  //  Enable pin 2 as INPUT for limit switch - set low
    PORTB |= (1 << 3);  //  Set Arduino pin 11 HIGH for pull up (encodPinB1)
    //pinMode(encodPinA1, INPUT); 
    //pinMode(encodPinB1, INPUT); 
    //digitalWrite(encodPinA1, HIGH);      // turns on pullup resistor
    //digitalWrite(encodPinB1, HIGH);
    PCICR  |= (1 << PCIE0);   // enable group interrupts on PORTB PCINT[7:0]
    PCMSK0 |= (1 << PCINT4); // enable interrupt pin 11 aka bit 3, try 10 for B, 11 forA
  sei();  // enable interrupts  bit 3=11 bit 2 = 10
  //attachInterrupt(1, rencoder, FALLING); 
  for(int i=0; i<NUMREADINGS; i++) readings[i] = 0;  // initialize readings to 0 
   //setup for PID
   Input = speed_act;     // reads encoder counts
   Setpoint = speed_req;  // reads the desired speed
   myPID.SetMode(AUTOMATIC);
  //pinMode(pwmPin, OUTPUT);
   analogWrite(PWM1, PWM_val);
   digitalWrite(InA1, LOW);  
   digitalWrite(InB1, HIGH);
}

void loop() {  
  getParam();                                          // check keyboard
  if((millis()-lastMilli) >= LOOPTIME)   {             // enter timed loop
    lastMilli = millis();
    getMotorData();                                    // calculate speed, volts and Amps
    PWM_val= updatePid(PWM_val, speed_req, speed_act); // compute PWM value     
    // 	This looks like the area for the PID lib       
    analogWrite(PWM1, PWM_val);  
    //myPID.Compute();    
    //Serial.print("OUTPUT: ");
    //Serial.println(Output);  
  }
    printMotorInfo();
    
}

void getMotorData()     // calculate speed, volts and Amps
{                                         
  static long countAnt = 0;  // last count			  // 64 pulses X 131 gear ratio = 8384, try 32 > 4192
  speed_act = ((count - countAnt)*(60*(1000/LOOPTIME)))/(32*131); // 16 pulses X 29 gear ratio = 464 counts per output shaft rev
  countAnt = count;   
  voltage = int(analogRead(Vpin) * 3.22 * 12.2/2.2);     // battery voltage: mV=ADC*3300/1024, voltage divider 10K+2K, was 12.2/2.2 for voltage
  //current = int(analogRead(Apin) * 3.22 * .77 *(1000.0/144.0));  // motor current - output: 130mV per Amp  132.0
  current = int(analogRead(Apin)*34);  //md.getM1CurrentMilliamps();
  //5.0 * .77 *(1000.0/144.0));analogRead(Apin) * 34;
  //// 5V / 1024 ADC counts / 144 mV per A = 34 mA per count
  current = digital_smooth(current, readings);                   // remove signal noise
}

int updatePid(int command, int targetValue, int currentValue)     // compute PWM value
{    
  float pidTerm = 0;                                              // PID correction
  int error=0;                                  
  static int last_error=0;                             
  error = abs(targetValue) - abs(currentValue); 
  pidTerm = (Kp * error) + (Kd * (error - last_error));                            
  last_error = error;
  //Setpoint = command;
  //Input = currentValue;
 
  /***********************************
    PID set up data - return the output
    Setpoint = targetValue;
    Input = currentValue; //analogRead(0);
    myPID.Compute();
    return Output;
    return OUTPUT return constrain(command + int(Output),0,255);
  ************************************/   
  //return speed_req;
  // return constrain(command + int(Output),0,255);
   return constrain(command + int(pidTerm), 0, 255);
}

void printMotorInfo()  {                      // display data
  
  if((millis()-lastMilliPrint) >= 500)   {                     
    lastMilliPrint = millis();
    Serial.print("SP:");             Serial.print(speed_req);  
    Serial.print("  RPM:");          Serial.print(speed_act);
    Serial.print("  PWM:");          Serial.print(PWM_val);  
    Serial.print("  V:");            Serial.print(float(voltage)/1000,1);
    Serial.print("  mA:");           Serial.println(current);

    if (current > CURRENT_LIMIT)            Serial.println("*** CURRENT_LIMIT ***");                
    if (voltage > 1000 && voltage*11 < LOW_BAT)   Serial.println("*** LOW_BAT ***");                
  }
}

//ISR for Pin 8 was 0 and 1, change to 3 and 2 for pins 11 and 10
ISR (PCINT0_vect)    // handle pin change interrupt for D8-D13;  triggered by 8
{	 //3	was 2 and 3	//2
if(bitRead(PINB,4)== bitRead(PINB,3)) {  //(PINB, x) is the syntax to read the input state from PORTB, bit x 
    //encoderValue++;
    count++;     
  } else {
    //encoderValue--;
    count--;    
  } 
}

void rencoder()  {          // pulse and direction, direct port reading to save cycles
  // this does not work
  if (PINB & 0b00000100) 
  //if(digitalRead(encodPinB1)==HIGH)
    count++;                // if(digitalRead(encodPinB1)==HIGH)   count ++;
  else 
    count--;                // if (digitalRead(encodPinB1)==LOW)   count --;
}

/*if (digitalRead(encoderPinA) == digitalRead(encoderPinB)) {
    encoderValue++; 
  } else {
    encoderValue--; */

int getParam()  {
  char param, cmd;
  if(!Serial.available())    return 0;
  delay(10);                  
  param = Serial.read();                            // get parameter byte
  if(!Serial.available())    return 0;
  cmd = Serial.read();                              // get command byte
  Serial.flush();
  switch (param) {
    case 'v':                                       // adjust speed
      if(cmd=='+')  {
        speed_req += 12;
        if(speed_req>400) speed_req=400;
      }
      if(cmd=='-')    {
        speed_req -= 12;
        if(speed_req<0)   speed_req=0;       
      }
      break;
    case 's':                                       // adjust direction
      if(cmd=='+'){
        digitalWrite(InA1, HIGH);
        digitalWrite(InB1, LOW);
      }
      if(cmd=='-')   {
        digitalWrite(InA1, LOW);
        digitalWrite(InB1, HIGH);
      }
      break;
    case 'o':                                       // user should type "oo"
      digitalWrite(InA1, LOW);
      digitalWrite(InB1, LOW);
      speed_req = 0;
      break;
    default: 
      Serial.println("???");
    }
}

int digital_smooth(int value, int *data_array)  {    // remove signal noise
  static int ndx=0;                                                         
  static int count=0;                          
  static int total=0;                          
  total -= data_array[ndx];               
  data_array[ndx] = value;                
  total += data_array[ndx];               
  ndx = (ndx+1) % NUMREADINGS;                                
  if(count < NUMREADINGS) count++;
  return total/count;
}
